@@ -89,7 +89,7 @@ class LibpqReader:
     app = None
 
     _url_pattern = (
-        "https://raw.githubusercontent.com/postgres/postgres/REL_{ver}_STABLE"
+        "https://raw.githubusercontent.com/postgres/postgres/{branch}"
         "/doc/src/sgml/libpq.sgml"
     )
 
@@ -130,14 +130,20 @@ class LibpqReader:
 
     @property
     def sgml_url(self):
-        return self._url_pattern.format(ver=self.version)
+        return self._url_pattern.format(branch=self.branch)
+
+    @property
+    def branch(self):
+        if self.version == "devel":
+            return "master"
+        return f"REL_{self.version}_STABLE"
 
     @property
     def version(self):
         return self.app.config.libpq_docs_version
 
 
-@lru_cache()
+@lru_cache
 def get_reader():
     return LibpqReader()
 
