@@ -4,18 +4,19 @@ Mappings between PostgreSQL and Python encodings.
 
 # Copyright (C) 2020 The Psycopg Team
 
+from __future__ import annotations
+
 import re
 import string
 import codecs
-from typing import Any, Dict, Optional, TYPE_CHECKING
+from typing import Any, TYPE_CHECKING
 
 from .pq._enums import ConnStatus
 from .errors import NotSupportedError
 from ._compat import cache
 
 if TYPE_CHECKING:
-    from .pq.abc import PGconn
-    from .connection import BaseConnection
+    from ._connection_base import BaseConnection
 
 OK = ConnStatus.OK
 
@@ -67,7 +68,7 @@ _py_codecs = {
     "WIN874": "cp874",
 }
 
-py_codecs: Dict[bytes, str] = {}
+py_codecs: dict[bytes, str] = {}
 py_codecs.update((k.encode(), v) for k, v in _py_codecs.items())
 
 # Add an alias without underscore, for lenient lookups
@@ -78,30 +79,13 @@ py_codecs.update(
 pg_codecs = {v: k.encode() for k, v in _py_codecs.items()}
 
 
-def conn_encoding(conn: "Optional[BaseConnection[Any]]") -> str:
+def conn_encoding(conn: BaseConnection[Any] | None) -> str:
     """
     Return the Python encoding name of a psycopg connection.
 
     Default to utf8 if the connection has no encoding info.
     """
-    if not conn or conn.closed:
-        return "utf-8"
-
-    pgenc = conn.pgconn.parameter_status(b"client_encoding") or b"UTF8"
-    return pg2pyenc(pgenc)
-
-
-def pgconn_encoding(pgconn: "PGconn") -> str:
-    """
-    Return the Python encoding name of a libpq connection.
-
-    Default to utf8 if the connection has no encoding info.
-    """
-    if pgconn.status != OK:
-        return "utf-8"
-
-    pgenc = pgconn.parameter_status(b"client_encoding") or b"UTF8"
-    return pg2pyenc(pgenc)
+    return conn.pgconn._encoding if conn else "utf-8"
 
 
 def conninfo_encoding(conninfo: str) -> str:
@@ -117,7 +101,7 @@ def conninfo_encoding(conninfo: str) -> str:
     pgenc = params.get("client_encoding")
     if pgenc:
         try:
-            return pg2pyenc(pgenc.encode())
+            return pg2pyenc(str(pgenc).encode())
         except NotSupportedError:
             pass
 
@@ -135,7 +119,7 @@ def py2pgenc(name: str) -> bytes:
 
 @cache
 def pg2pyenc(name: bytes) -> str:
-    """Convert a Python encoding name to PostgreSQL encoding name.
+    """Convert a PostgreSQL encoding name to Python encoding name.
 
     Raise NotSupportedError if the PostgreSQL encoding is not supported by
     Python.
