@@ -1,7 +1,9 @@
+from __future__ import annotations
+
 import asyncio
 import selectors
 import sys
-from typing import Any, Dict, List
+from typing import Any
 
 import pytest
 
@@ -30,6 +32,7 @@ def pytest_configure(config):
         "gevent: the test requires the gevent module to be installed",
         "dns: the test requires dnspython to run",
         "postgis: the test requires the PostGIS extension to run",
+        "numpy: the test requires numpy module to be installed",
     ]
 
     for marker in markers:
@@ -70,9 +73,11 @@ def pytest_sessionstart(session):
     cache.set("segfault", True)
 
 
-asyncio_options: Dict[str, Any] = {}
-if sys.platform == "win32" and sys.version_info >= (3, 8):
-    asyncio_options["policy"] = asyncio.WindowsSelectorEventLoopPolicy()
+asyncio_options: dict[str, Any] = {}
+if sys.platform == "win32":
+    asyncio_options["loop_factory"] = (
+        asyncio.WindowsSelectorEventLoopPolicy().new_event_loop
+    )
 
 
 @pytest.fixture(
@@ -86,7 +91,7 @@ def anyio_backend(request):
     return backend, options
 
 
-allow_fail_messages: List[str] = []
+allow_fail_messages: list[str] = []
 
 
 def pytest_sessionfinish(session, exitstatus):
