@@ -1,5 +1,7 @@
+from __future__ import annotations
+
+import sys
 import pickle
-from typing import List
 from weakref import ref
 
 import pytest
@@ -8,10 +10,22 @@ import psycopg
 from psycopg import pq
 from psycopg import errors as e
 
-from .utils import eur, gc_collect
+from .utils import eur
 from .fix_crdb import is_crdb
 
 
+def test_finishedpgconn(pgconn):
+    with pytest.raises(TypeError):
+        e.FinishedPGconn.connect()
+
+    assert pgconn.socket
+    finished = e.finish_pgconn(pgconn)
+    with pytest.raises(e.OperationalError, match="connection is closed"):
+        finished.socket
+    with pytest.raises(e.OperationalError, match="connection is closed"):
+        pgconn.socket
+
+
 @pytest.mark.crdb_skip("severity_nonlocalized")
 def test_error_diag(conn):
     cur = conn.cursor()
@@ -37,7 +51,7 @@ def test_diag_right_attr(pgconn, monkeypatch):
     diag = e.Diagnostic(res)
 
     to_check: pq.DiagnosticField
-    checked: List[pq.DiagnosticField] = []
+    checked: list[pq.DiagnosticField] = []
 
     def check_val(self, v):
         nonlocal to_check
@@ -170,7 +184,11 @@ def test_diag_pickle(conn):
 
 
 @pytest.mark.slow
-def test_diag_survives_cursor(conn):
+@pytest.mark.xfail(
+    (pq.__impl__ in ("c", "binary") and sys.version_info[:2] == (3, 12)),
+    reason="Something with Exceptions, C, Python 3.12",
+)
+def test_diag_survives_cursor(conn, gc_collect):
     cur = conn.cursor()
     with pytest.raises(e.Error) as exc:
         cur.execute("select * from nosuchtable")
@@ -216,7 +234,6 @@ def test_diag_from_commit(conn):
     assert exc.value.diag.sqlstate == "23503"
 
 
-@pytest.mark.asyncio
 @pytest.mark.crdb_skip("deferrable")
 async def test_diag_from_commit_async(aconn):
     cur = aconn.cursor()
@@ -307,3 +324,32 @@ def test_pgresult_pickle(conn):
 
 def test_blank_sqlstate(conn):
     assert e.get_base_exception("") is e.DatabaseError
+
+
+@pytest.mark.parametrize(
+    "msg",
+    [
+        'connection to server at "2001:1488:fffe:20::229", port 5432 failed',
+        "HORROR: foo\n",
+    ],
+)
+def test_strip_severity_unstripped(msg):
+    from psycopg.pq.misc import strip_severity
+
+    out = strip_severity(msg)
+    assert out == msg.strip()
+
+
+@pytest.mark.parametrize(
+    "msg",
+    [
+        "ERROR: foo\n",
+        "ERRORE: foo\nbar\n",
+        "오류: foo: bar",
+    ],
+)
+def test_strip_severity_l10n(msg):
+    from psycopg.pq.misc import strip_severity
+
+    out = strip_severity(msg)
+    assert out == msg.split(":", 1)[1].strip()
