@@ -1,5 +1,8 @@
+from __future__ import annotations
+
 import gc
-from typing import List, Any
+from math import prod
+from typing import Any
 from decimal import Decimal
 
 import pytest
@@ -9,10 +12,10 @@ from psycopg import pq
 from psycopg import sql
 from psycopg.adapt import PyFormat, Transformer, Dumper
 from psycopg.types import TypeInfo
-from psycopg._compat import prod
 from psycopg.postgres import types as builtins
 from psycopg.types.array import register_array
 
+from ..test_adapt import StrNoneDumper, StrNoneBinaryDumper
 
 tests_str = [
     ([[[[[["a"]]]]]], "{{{{{{a}}}}}}"),
@@ -49,6 +52,16 @@ def test_dump_list_str(conn, obj, want, fmt_in):
     assert cur.fetchone()[0]
 
 
+@pytest.mark.parametrize("fmt_in", PyFormat)
+def test_dump_list_str_none(conn, fmt_in):
+    cur = conn.cursor()
+    cur.adapters.register_dumper(str, StrNoneDumper)
+    cur.adapters.register_dumper(str, StrNoneBinaryDumper)
+
+    cur.execute(f"select %{fmt_in.value}::text[]", (["foo", "", "bar"],))
+    assert cur.fetchone()[0] == ["foo", None, "bar"]
+
+
 @pytest.mark.parametrize("fmt_out", pq.Format)
 def test_load_empty_list_str(conn, fmt_out):
     cur = conn.cursor(binary=fmt_out)
@@ -295,7 +308,7 @@ def test_load_array_no_comma_separator(conn):
 @pytest.mark.parametrize("fmt_out", pq.Format)
 def test_load_nested_array(conn, fmt_out):
     dims = [3, 4, 5, 6]
-    a: List[Any] = list(range(prod(dims)))
+    a: list[Any] = list(range(prod(dims)))
     for dim in dims[-1:0:-1]:
         a = [a[i : i + dim] for i in range(0, len(a), dim)]
 
